//
// LDASTools diskcache - Tools for querying a collection of files on disk
//
// Copyright (C) 2022 California Institute of Technology
//
// LDASTools diskcache is free software; you may redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 (GPLv2) of the
// License or at your discretion, any later version.
//
// LDASTools diskcache is distributed in the hope that it will be useful, but
// without any warranty or even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public
// License (GPLv2) for more details.
//
// Neither the names of the California Institute of Technology (Caltech),
// The Massachusetts Institute of Technology (M.I.T), The Laser
// Interferometer Gravitational-Wave Observatory (LIGO), nor the names
// of its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// You should have received a copy of the licensing terms for this
// software included in the file LICENSE located in the top-level
// directory of this package. If you did not, you can view a copy at
// http://dcc.ligo.org/M1500244/LICENSE
//

#include <diskcache_config.h>

#include <unordered_set>

#include "genericAPI/Logging.hh"
#include "genericAPI/LDASplatform.hh"

#include "diskcacheAPI/Common/Variables.hh"

#include "diskcacheAPI/Cache/HotDirectory.hh"

#include "diskcacheAPI/Streams/ASCII.hh"
#include "diskcacheAPI/Streams/Binary.hh"
#include "diskcacheAPI/Streams/FStream.hh"

#include "Cache/Devices.hh"

#include "Commands.hh"
#include "diskcachecmd.hh"
#include "IO.hh"
#include "MetaCommands.hh"
#include "MountPointScanner.hh"

using diskCache::Commands::updateFileExtList;
using diskCache::Commands::updateMountPtList;

namespace diskCache
{
    namespace MetaCommand
    {
        //===================================================================
        // Scan::Config_
        //===================================================================
        Scan::Config_::Config_( Scan& Command )
            : command( Command ), state( NON_BLOCK )
        {
        }

        void
        Scan::Config_::Parse( std::istream& Stream )
        {
            base_type::Parse( Stream );
        }

        void
        Scan::Config_::ParseBlock( const std::string& Value )
        {
            if ( Value.compare( "DAEMON" ) == 0 )
            {
                state = BLOCK_DAEMON;
            }
            else if ( Value.compare( "EXTENSIONS" ) == 0 )
            {
                state = BLOCK_EXTENSIONS;
            }
            else if ( Value.compare( "EXCLUDED_DIRECTORIES" ) == 0 )
            {
                state = BLOCK_EXCLUDED_DIRECTORIES;
            }
            else if ( Value.compare( "EXCLUDED_PATTERN" ) == 0 )
            {
                command.reset_excluded_patterns( );
                state = BLOCK_EXCLUDED_PATTERNS;
            }
            else if ( Value.compare( "MOUNT_POINTS" ) == 0 )
            {
                state = BLOCK_MOUNT_POINTS;
            }
            else
            {
                //---------------------------------------------------------------
                // Tried everything we know. Flag as unknown block
                //---------------------------------------------------------------
                state = BLOCK_UNKNOWN;
            }
        }

        void
        Scan::Config_::ParseKeyValue( const std::string& Key,
                                      const std::string& Value )
        {
            switch ( state )
            {
            case NON_BLOCK:
            case BLOCK_DAEMON:
                try
                {
                    diskCache::Common::Variables::Cache( Key, Value );
                    return;
                }
                catch ( const std::invalid_argument& Error )
                {
                    //-------------------------------------------------------------
                    // This is the case where the variable is NOT handled by
                    //   the generic routine.
                    //-------------------------------------------------------------
                }
                //-----------------------------------------------------------------
                // These entries should be ignored for the scan subcommand
                //-----------------------------------------------------------------
                static std::unordered_set< std::string > ignored_keys(
                    { "CACHE_WRITE_DELAY_SECS",
                      "LOG",
                      "LOG_DIRECTORY",
                      "LOG_ARCHIVE_DIRECTORY",
                      "LOG_DEBUG_LEVEL",
                      "LOG_FORMAT",
                      "LOG_ROTATE_ENTRY_COUNT",
                      "SCAN_INTERVAL",
                      "SERVER_PORT" } );

                if ( ignored_keys.find( Key ) != ignored_keys.end( ) )
                {
                    return;
                }

                if ( Key.compare( "OUTPUT_ASCII" ) == 0 )
                {
                    command.set_output_ascii( Value );
                }
                else if ( Key.compare( "OUTPUT_BINARY" ) == 0 )
                {
                    command.set_output_binary( Value );
                }
                else if ( Key.compare( "OUTPUT_BINARY_VERSION" ) == 0 )
                {
                    command.set_output_binary_version( Value );
                }
                else
                {
                    if ( state == BLOCK_DAEMON )
                    {
                        //-----------------------------------------------------------
                        /// \todo
                        ///     Need to produce and exception regarding unknown
                        ///     keyword found
                        //-----------------------------------------------------------
                    }
                }
                break;
            case BLOCK_EXCLUDED_DIRECTORIES:
            case BLOCK_EXCLUDED_PATTERNS:
            case BLOCK_EXTENSIONS:
            case BLOCK_MOUNT_POINTS:
                //---------------------------------------------------------------
                /// \todo
                ///     Need to produce an exception regarding parse error
                ///     in configuration file as a key/value pair was
                ///     encountered where a word value was expected.
                //---------------------------------------------------------------
                break;
            default:
                //---------------------------------------------------------------
                // Ignore all other cases
                //---------------------------------------------------------------
                break;
            }
        }

        void
        Scan::Config_::ParseWord( const std::string& Value )
        {
            if ( Value.size( ) <= 0 )
            {
                //---------------------------------------------------------------
                // Ignore empty words
                //---------------------------------------------------------------
                return;
            }
            //-----------------------------------------------------------------
            // Process the word according to the block it was found in
            //-----------------------------------------------------------------
            switch ( state )
            {
            case BLOCK_EXTENSIONS:
                command.push_extension( Value );
                break;
            case BLOCK_EXCLUDED_DIRECTORIES:
                command.push_excluded_directory( Value );
                break;
            case BLOCK_EXCLUDED_PATTERNS:
                command.push_excluded_pattern( Value );
                break;
            case BLOCK_MOUNT_POINTS:
                command.push_mount_point( Value );
                break;
            case BLOCK_DAEMON:
            case BLOCK_UNKNOWN:
            case NON_BLOCK:
                // This shouild never happen.
                /// \todo
                /// Have Scan::Config_::ParseWord throw exception
                /// if it reaches an unreachable state.
                break;
            }
        }

        //===================================================================
        // Scan
        //===================================================================
        OptionSet& Scan::m_options( Scan::init_options( ) );

        OptionSet&
        Scan::init_options( )
        {
            static OptionSet retval;

            retval.Synopsis( "Subcommand: scan" );

            retval.Summary( "The scan sub command is intended to"
                            " scan a set of directories for files of interest"
                            " and generate a memory cache image." );

            retval.Add( Option( OPT_CONCURRENCY,
                                "concurrency",
                                Option::ARG_REQUIRED,
                                "Number of mount points to scan concurrently",
                                "integer" ) );

            retval.Add( Option(
                OPT_CONFIGURATION_FILE,
                "configuration-file",
                Option::ARG_REQUIRED,
                "Name of file containing additional configuration information.",
                "filename" ) );

            retval.Add( Option( OPT_EXTENSIONS,
                                "extensions",
                                Option::ARG_REQUIRED,
                                "Comma seperated list of file extensions",
                                "list" ) );

            retval.Add( Option( OPT_MOUNT_POINTS,
                                "mount-points",
                                Option::ARG_REQUIRED,
                                "Comma seperated list of mount points to scan",
                                "list" ) );

            retval.Add( Option( OPT_OUTPUT_ASCII,
                                "output-ascii",
                                Option::ARG_REQUIRED,
                                "Filename for the ascii output; '-' to direct "
                                "to standard output",
                                "filename" ) );

            retval.Add( Option( OPT_OUTPUT_BINARY,
                                "output-binary",
                                Option::ARG_REQUIRED,
                                "Filename for the binary output",
                                "filename" ) );

            retval.Add(
                Option( OPT_TYPE,
                        "type",
                        Option::ARG_REQUIRED,
                        "Type pattern to use for search. (Default: all)",
                        "pattern" ) );

            retval.Add(
                Option( OPT_VERSION_ASCII,
                        "version-ascii",
                        Option::ARG_REQUIRED,
                        "Version of the ascii diskcache dump format to output",
                        "version" ) );
            retval.Add(
                Option( OPT_VERSION_BINARY,
                        "version-binary",
                        Option::ARG_REQUIRED,
                        "Version of the binary diskcache dump format to output",
                        "version" ) );
            return retval;
        }

        Scan::Scan( CommandLineOptions& Args,
                    const std::string&  DefaultConfigurationFilename )
            : m_args( Args ), m_output_ascii( "" ), m_output_binary( "" ),
              m_version_ascii( VERSION_DEFAULT_ASCII ),
              m_version_binary( VERSION_DEFAULT_BINARY )
        {
            {
                configuration_filename = DefaultConfigurationFilename;
                std::ifstream stream( configuration_filename.c_str( ) );

                if ( stream.is_open( ) )
                {
                    Config_ cfg( *this );

                    cfg.Parse( stream );
                }
            }
            if ( m_args.empty( ) == false )
            {
                //---------------------------------------------------------------
                // Parse the commands
                //---------------------------------------------------------------
                std::string arg_name;
                std::string arg_value;
                bool        parsing( true );

                while ( parsing )
                {
                    switch ( m_args.Parse( m_options, arg_name, arg_value ) )
                    {
                    case CommandLineOptions::OPT_END_OF_OPTIONS:
                        parsing = false;
                        break;
                    case OPT_CONCURRENCY:
                    {
                        std::istringstream value_stream( arg_value );

                        INT_4U concurrency;

                        value_stream >> concurrency;

                        ScanConcurrency( concurrency );
                    }
                    break;
                    case OPT_CONFIGURATION_FILE:
                    {
                        configuration_filename = arg_value;
                        std::ifstream stream( configuration_filename.c_str( ) );

                        if ( stream.is_open( ) )
                        {
                            Config_ cfg( *this );

                            cfg.Parse( stream );
                        }
                    }
                    break;
                    case OPT_EXTENSIONS:
                    {
                        //-----------------------------------------------------------
                        // Generate list of extensions
                        //-----------------------------------------------------------
                        size_t pos = 0;
                        size_t end = 0;

                        m_extensions.resize( 0 );
                        while ( end != std::string::npos )
                        {
                            end = arg_value.find_first_of( ",", pos );

                            push_extension(
                                arg_value.substr( pos,
                                                  ( ( end == std::string::npos )
                                                        ? end
                                                        : end - pos ) ) );
                            pos = end + 1;
                        }
                    }
                    break;
                    case OPT_MOUNT_POINTS:
                    {
                        //-----------------------------------------------------------
                        // Generate list of mount points
                        //-----------------------------------------------------------
                        size_t pos = 0;
                        size_t end = 0;

                        while ( end != std::string::npos )
                        {
                            end = arg_value.find_first_of( ",", pos );
                            m_mount_points.push_back(
                                arg_value.substr( pos,
                                                  ( ( end == std::string::npos )
                                                        ? end
                                                        : end - pos ) ) );
                            pos = end + 1;
                        }
                    }
                    break;
                    case OPT_TYPE:
                        m_type = arg_value;
                        break;
                    case OPT_OUTPUT_ASCII:
                        m_output_ascii = arg_value;
                        break;
                    case OPT_OUTPUT_BINARY:
                        m_output_binary = arg_value;
                        break;
                    case OPT_VERSION_ASCII:
                    {
                        std::istringstream version_stream( arg_value );

                        version_stream >> std::hex >> m_version_ascii >>
                            std::dec;
                    }
                    break;
                    case OPT_VERSION_BINARY:
                    {
                        std::istringstream version_stream( arg_value );

                        version_stream >> std::hex >> m_version_binary >>
                            std::dec;
                    }
                    break;
                    default:
                        break;
                    }
                }
            }
        }

        const OptionSet&
        Scan::Options( )
        {
            return m_options;
        }

        void
        Scan::operator( )( )
        {
            //-----------------------------------------------------------------
            // Add the extensions
            //-----------------------------------------------------------------
            updateFileExtList( m_extensions );

            //-----------------------------------------------------------------
            // Add the mount points
            //-----------------------------------------------------------------
            if ( m_mount_points.size( ) > 0 )
            {
                MountPointManagerSingleton::UpdateResults status;

                updateMountPtList( status, m_mount_points, false );
            }

            //-----------------------------------------------------------------
            // Scan the directories
            //-----------------------------------------------------------------

            {
                diskCache::MountPointScanner::ScanResults results;
                ScanMountPointList( results );

                /// \todo format results for output to stream
                /// diskCache::ASCII::Translate( &std::cout, results );
            }

            //-----------------------------------------------------------------
            // Write the scanned data
            //-----------------------------------------------------------------
            if ( m_output_ascii.empty( ) ||
                 ( m_output_ascii.compare( "-" ) == 0 ) )
            {
                //-------------------------------------------------------------
                // Dump the default mount point manager
                //-------------------------------------------------------------
                diskCache::Streams::OASCII stream( std::cout, m_version_ascii );

                diskCache::Write( stream );
            }
            else if ( m_output_ascii.empty( ) == false )
            {
                //-------------------------------------------------------------
                /// \todo
                ///     When the output is ascii, the query options should allow
                ///     the user to request a subset of the entire cache.
                //-------------------------------------------------------------
                //-------------------------------------------------------------
                // Dump the default mount point manager
                //-------------------------------------------------------------
                diskCache::Streams::OFStream file_stream( m_output_ascii );
                diskCache::Streams::OASCII   stream( file_stream,
                                                   m_version_ascii );

                diskCache::Write( stream );
            }
            if ( m_output_binary.empty( ) == false )
            {
                //-------------------------------------------------------------
                /// \todo
                ///     When the output is binary, the query options should
                ///     allow the user to request a subset of the entire cache.
                //-------------------------------------------------------------
                //-------------------------------------------------------------
                // Dump the default mount point manager
                //-------------------------------------------------------------
                diskCache::Streams::OFStream file_stream( m_output_binary );
                diskCache::Streams::OBinary  stream( file_stream,
                                                    m_version_binary );

                diskCache::Write( stream );
            }
        }

        inline void
        Scan::push_excluded_directory( const std::string& Directory )
        {
            excluded_directories.push_back( Directory );
        }

        inline void
        Scan::push_excluded_pattern( const std::string& Pattern )
        {
            excluded_patterns.push_back( Pattern );
        }

        inline void
        Scan::push_extension( const std::string& Extension )
        {
            m_extensions.push_back( Extension );
        }

        inline void
        Scan::push_mount_point( const std::string& MountPoint )
        {
            m_mount_points.push_back( MountPoint );
        }

        inline void
        Scan::set_output_ascii( const std::string& Value )
        {
            m_output_ascii = Value;
        }

        inline void
        Scan::set_output_binary( const std::string& Value )
        {
            m_output_binary = Value;
        }

        inline void
        Scan::set_output_binary_version( const std::string& Value )
        {
            if ( Value.substr( 0, 2 ).compare( "0x" ) == 0 )
            {
                std::istringstream s( Value.substr( 2 ) );
                s >> std::hex >> m_version_binary;
            }
            else if ( Value[ 0 ] == '0' )
            {
                std::istringstream s( Value.substr( 1 ) );
                s >> std::oct >> m_version_binary;
            }
            else
            {
                std::istringstream s( Value );
                s >> std::hex >> m_version_binary;
            }
        }

        inline void
        Scan::reset_excluded_patterns( )
        {
            excluded_patterns.resize( 0 );
        }

    } /* namespace MetaCommand */
} /* namespace diskCache */
